from typing_extensions import Unpack, NotRequired, TypedDict, Mapping
from dslog import Logger
from kv.api import KV
import moveread.pipelines.preprocess as pre
from .spec import Game
from .spec_codegen import Workflow
from .pipelines import preinput, join

class Params(TypedDict):
  games: KV[Game]
  imgGameIds: KV[str]
  received_imgs: KV[Mapping[str, pre.Result]]
  images: KV[bytes]
  images_path: NotRequired[str | None]

class Artifacts(pre.Artifacts):
  ...

def artifacts(**queues: Unpack[Workflow.InternalQueues]):
  
  def _bound(
    *, logger: Logger = Logger.click().prefix('[GAME PREPROCESS]'),
    images: KV[bytes], images_path: str | None = None,
    games: KV[Game], imgGameIds: KV[str],
    received_imgs: KV[Mapping[str, pre.Result]]
  ) -> Artifacts:
    
    artifacts = pre.Workflow.artifacts(**queues['preprocess']['internal'])(
      logger=logger.prefix('[PREPROCESS]'), images=images,
      images_path=images_path
    )

    artifacts.processes = { f'preprocess-{id}': ps for id, ps in artifacts.processes.items() }

    artifacts.processes['preinput'] = preinput(
      logger=logger.prefix('[PREINPUT]'), games=games, imgGameIds=imgGameIds,
      **queues['preinput']
    )

    artifacts.processes['join'] = join(
      logger=logger.prefix('[JOIN]'), games=games, imgGameIds=imgGameIds,
      received_imgs=received_imgs, **queues['join']
    )

    return Artifacts(processes=artifacts.processes, api=artifacts.api)
  
  return _bound


from typing import Sequence
from pydantic import BaseModel
from pipeteer import Pipeline, Workflow
from scoresheet_models import ModelID
import moveread.pipelines.preprocess as pre
Preprocess = pre.Workflow

PreInput = pre.Input
PreInput.__name__ = 'PreInput'

PreResult = pre.Result
PreResult.__name__ = 'PreResult'

class Input(BaseModel):
  model: ModelID 
  imgs: Sequence[str]

class Game(BaseModel):
  model: ModelID
  imgIds: Sequence[str]

class Result(BaseModel):
  preprocessed_imgs: Sequence[PreResult]

workflow = Workflow[Input, Result](
  'preinput', Result,
  pipelines=dict(
    preinput=Pipeline(Input, PreInput),
    preprocess=pre.workflow,
    join=Pipeline(PreResult, Result)
  )
)

def codegen():
  workflow.codegen(__file__, overwrite=True)



import asyncio
from haskellian import either as E
from dslog import Logger
from kv.api import KV
from ..spec import Game, PreInput
from ..spec_codegen import PreinputPipeline as Preinput

async def preinput(
  Qin: Preinput.QueueIn, Qout: Preinput.QueueOut,
  *, games: KV[Game], imgGameIds: KV[str],
  logger = Logger.rich().prefix('[PREINPUT]')
):
  @E.do()
  async def input_one():
    gameId, task = (await Qin.read()).unsafe()
    imgIds = [f'{gameId}-{i}' for i in range(len(task.imgs))]
    (await games.insert(gameId, Game(model=task.model, imgIds=imgIds))).unsafe()
    E.sequence(await asyncio.gather(*[
      imgGameIds.insert(imgId, gameId)
      for imgId in imgIds
    ])).unsafe()
    E.sequence(await asyncio.gather(*[
      Qout.push(imgId, PreInput(model=task.model, img=img))
      for imgId, img in zip(imgIds, task.imgs)
    ])).unsafe()
    (await Qin.pop(gameId)).unsafe()
    logger(f'Pushed tasks from "{gameId}"', level='DEBUG')

  while True:
    r = await input_one()
    if r.tag == 'left':
      logger(r.value, level='ERROR')
      await asyncio.sleep(1)